package com.xefer.rebel;

import java.io.*;
import java.lang.*;
import java.util.*;

import javax.servlet.*;
import javax.servlet.http.*;

import javax.xml.parsers.*;
import javax.xml.transform.*;
import javax.xml.transform.dom.*;
import javax.xml.transform.stream.*;

import org.apache.xpath.*;
import org.w3c.dom.*;

/**
* The base class of all XBEL elements
*/  

abstract class XBELElement {
  
  Element element;

  protected XBELElement() {
  
  }

  public XBELElement (Element element) {
    this.element = element;  
  }
  
  /**
   * Gets the id attribute of the underlying XBEL element 
   *
   * @return String the id of the given element
   */
  
  public String getId() {
  
    return element.getAttribute("id");
  
  }
  
  /**
   * Sets the id attribute of the underlying XBEL element 
   *
   * @param String the id of the given element being updated
   */
  
  public void setId(String id) {
  
    element.setAttribute("id", id);
  
  }
  
  /**
   * Gets the tagname of the underlying XBEL element 
   *
   * @return String the tag name of the given element
   */
   
  public String getType() {
  
    return element.getTagName();
  
  }
  
  /** 
   *
   * Delete removes self from parent
   *
   * @return void
   */
   
  public void delete() throws ResourceException {
  
    Node parentNode = element.getParentNode();
    parentNode.removeChild(element);     
   
  }
  
  /** 
   *
   * Handles the POSTing of a subordinate resource
   *
   * @param entity <code>XBELElement</code> the entity to be posted
   * @return void
   */
   
  public void post(XBELElement entity) throws ResourceException {
  
    NodeList nodes    = element.getChildNodes();
    
    String childTag = entity.getType();
    String childId  = entity.getId();
    
    // first see if we have a duplicate id of the same type
    int length = nodes.getLength();
    for(int i = 0; i < length; i++) {
    
      Node node = nodes.item(i);
      
      if(node.getNodeType() == Node.ELEMENT_NODE) {
      
        XBELElement child = XBELElementFactory.newElement((Element) node);
        
        String tagname = child.getType();
        String id      = child.getId();
        
        if(tagname.equals(childTag) && id.equals(childId)) {
        
          replaceChild(entity, child);
        
        }
        
      }
    
    }
    
    // if we make it through the loop without the condition being
    // met then we have a brand new child so just append it
    
    addChild(entity);
    
  }
  
  /**
   * putResource
   *
   * @param entity <code>XBELElement</code> the entity to store
   * @return void
   * 
   */
   
  public void put(XBELElement entity) throws ResourceException {

    if(! getType().equals(entity.getType())) {
    
      throw new ResourceException(HttpServletResponse.SC_CONFLICT, "Can't store <" + entity.getType() + "> at URL, check type");
      
    }

    if(! getId().equals(entity.getId())) {

      throw new ResourceException(HttpServletResponse.SC_CONFLICT, "The id of the supplied entity [" + entity.getId() + "] does not match the resource [" + getId() + "]");

    }

    //
    // replace the old node with the new node
    //
    
    replaceSelf(entity);
    
  }
  
  /**
   * Returns the resource Location URL based on the current url and the 
   * resource type
   */
   
  abstract public String getLocationUrl(StringBuffer url);
  
  /**
   * Returns true if this XBELElement is directly locatable through a URL.  F
   *
   * @returns boolean
   */
     
  abstract public boolean isLocatable();
  
  /** 
   * Replace the given child with the given entity
   *
   * @param entity XBELElement to replace the current element
   */
  protected void replaceSelf(XBELElement entity) {
    
    Document document = element.getOwnerDocument();
    Element  imported = (Element) document.importNode(entity.element, true);

    Node parentNode   = element.getParentNode();
    
    parentNode.replaceChild(imported, element);  
    
  }
  
  /** 
   * Replace on child element with another
   *
   * @param entity XBELElement of the old child
   * @param child XBELElement of the new child
   */
  protected void replaceChild(XBELElement entity, XBELElement child) {
  
    Document document = element.getOwnerDocument();
    Element  imported = (Element) document.importNode(entity.element, true);
    
    element.replaceChild(imported, child.element);
  
  }
  
  /** 
   * Replace the current resource with the given entity
   *
   * @param child XBELElement of the new child
   */
  protected void addChild(XBELElement entity) {
  
    Document document = element.getOwnerDocument();
    Element  imported = (Element) document.importNode(entity.element, true);
        
    element.appendChild(imported);
    entity.element = imported;
    
  }
  
  /** 
   * Replace the current resource with the given entity without first
   * importing it into the parent elements document
   *
   * @param child XBELElement of the new child
   */
  protected void addChildDirect(XBELElement entity) {
  
    // I need to figure out why this function is even necessary.
    // For some reason the elements added by the temp tree
    // aren't added correctly if I import into teh elements doc first
    element.appendChild(entity.element);
    
  }
  
  
  
  /**
   * Write out the element ot the given output stream
   * 
   * @param context ServletContext that this process is executing in
   * @param stream OutputStream of the to be written to
   */  
  public void write(ServletContext context, OutputStream stream) throws ResourceException {
  
    try {
    
      // since I'm just doing an identity xfrom, this shouldn't be necessary
      // todo: look into extra space issues...
      InputStream istream =  context.getResourceAsStream("/WEB-INF/identity.xsl");
      
      TransformerFactory tFactory = TransformerFactory.newInstance();
      Transformer identity = tFactory.newTransformer( new StreamSource(istream) );

      Source source = new DOMSource(element);
      Result result = new StreamResult(stream);

      identity.transform(source, result);

    } catch (Exception ex) {
    
      throw new ResourceException(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, "Unable to write out element with identity tranformation", ex);
    
    }
    
  }

}